import re
from MultiMapping import MultiMapping

NAME_RE = r"[a-zA-Z][a-zA-Z0-9_]*"
_parse_expr = re.compile(r"(%s):(.*)" % NAME_RE).match
_valid_name = re.compile('%s$' % NAME_RE).match

class TALESError(Exception):
    '''TALES Error'''

class RegistrationError(TALESError):
    '''TALES Type Registration Error'''

class Iterator:
    '''Simple Iterator class for use in Contexts'''
    def __init__(self, name, seq, context):
        self.name = name
        self.seq = seq
        self._context = context
        self.nextIndex = 0

    def next(self):
        i = self.nextIndex
        try:
            item = self.seq[i]
        except IndexError:
            return 0
        self.index = i
        self.nextIndex = i+1
        self._context.setLocal(self.name, item)
        return 1

    def number(self): return self.nextIndex

    def sequence_start(self): return self.nextIndex == 1

    def sequence_end(self):
        try: self.seq[self.nextIndex]
        except IndexError: return 1
        return 0

class Engine:
    '''Expression Engine

    An instance of this class keeps a mutable collection of expression
    type handlers.  It can compile expression strings by delegating to
    these handlers.  It can provide an expression Context, which is
    capable of holding state and evaluating compiled expressions.
    '''
    Iterator = Iterator

    def __init__(self, Iterator=None):
        self.types = {}
        if Iterator is not None:
            self.Iterator = Iterator

    def registerType(self, name, handler):
        if not _valid_name(name):
            raise RegistrationError, 'Invalid Expression type "%s".' % name
        types = self.types
        if types.has_key(name):
            raise RegistrationError, (
                'Multiple registrations for Expression type "%s".' %
                name)
        types[name] = handler

    def getTypes(self):
        return self.types

    def compile(self, expression):
        m = _parse_expr(expression)
        if m:
            type, expr = m.group(1, 2)
        else:
            type = "standard"
            expr = expression
        try:
            handler = self.types[type]
        except KeyError:
            raise TALESError, (
                'Unrecognized expression type "%s".' % type)
        try:
            return handler(type, expr, self)
        except TypeError:
            return handler(type, expr)
    
    def getContext(self, **contexts):
        return Context(self, contexts)


class Context:
    '''Expression Context

    An instance of this class holds context information that it can
    use to evaluate compiled expressions.
    '''

    def __init__(self, engine, contexts):
        self._engine = engine
        self.contexts = contexts
        # Keep track of what contexts get pushed as each scope begins.
        self._ctxts_pushed = []
        # These contexts will need to be pushed if they are true (non-empty).
        self._current_ctxts = {'local': 1, 'loop': 1}
        contexts['local'] = lv = MultiMapping()
        init_local = contexts.get('local', None)
        if init_local:
            lv.push(init_local)
        contexts['global'] = gv = contexts.get('global', {})
        contexts['var'] = MultiMapping(gv, lv)
        contexts['loop'] = MultiMapping()
        
    def beginScope(self):
        pushed = []
        ctxts = self._current_ctxts
        for ctxname, ctx in ctxts.items():
            if ctx:
                # Push a non-empty scoped context, and make a fresh one.
                ctxts[ctxname] = ctx = {}
                self.contexts[ctxname].push(ctx)
                pushed.append(ctxname)
        # Remember what we pushed.
        self._ctxts_pushed.append(pushed)

    def endScope(self):
        assert len(self._ctxts_pushed), (
            "more endScope() than beginScope() calls")
        pushed = self._ctxts_pushed.pop()
        ctxts = self._current_ctxts
        # Clear all scoped contexts
        #for k in ctxts.keys():
        #    ctxts[k] = {}
        # Pop the ones that were pushed at the beginning of the scope.
        for ctxname in pushed:
            #XXX This is silly
            mm=self.contexts[ctxname]
            mm.pop() # toss top of namespace stack
            try:
                # now try to get new top
                m=mm.pop()
            except IndexError:
                # Nothing there, put marker in current context
                ctxts[ctxname] = 1
            else:
                # OK, capture the current context and push it back
                ctxts[ctxname] = m
                mm.push(m)
                
            # ctxts[ctxname] = self.contexts[ctxname].pop()

    def setLocal(self, name, value):
        self._current_ctxts['local'][name] = value

    def setGlobal(self, name, value):
        self.contexts['global'][name] = value

    def setRepeat(self, name, expr):
        expr = self.evaluate(expr)
        it = self._engine.Iterator(name, expr, self)
        self._current_ctxts['loop'][name] = it
        return it

    def evaluate(self, expression):
        if type(expression) is type(''):
            expression = self._engine.compile(expression)
        return expression(self)

    evaluateValue = evaluate

    def evaluateBoolean(self, expr):
        return not not self.evaluate(expr)

    def evaluateText(self, expr):
        text = self.evaluate(expr)
        if text is not None:
            text = str(text)
        return text

    def evaluateStructure(self, expr):
        # XXX Should return None or a DOM tree
        return self.evaluate(expr)

    def evaluateMacro(self, expr):
        # XXX Should return None or a macro definition
        return self.evaluate(expr)


class SimpleExpr:
    '''Simple example of an expression type handler'''
    def __init__(self, name, expr):
        self._name = name
        self._expr = expr
    def __call__(self, econtext):
        return self._name, self._expr
    def __repr__(self):
        return '<SimpleExpr %s %s>' % (self._name, `self._expr`)






